package org.jboss.security.plugins;

import java.lang.reflect.Method;
import java.lang.reflect.UndeclaredThrowableException;
import java.security.Principal;
import java.security.acl.Group;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import javax.security.auth.Subject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;
import org.jboss.logging.Logger;
import org.jboss.security.AnybodyPrincipal;
import org.jboss.security.NobodyPrincipal;
import org.jboss.security.RealmMapping;
import org.jboss.security.SecurityAssociation;
import org.jboss.security.SubjectSecurityManager;
import org.jboss.security.auth.callback.SecurityAssociationHandler;
import org.jboss.security.plugins.SubjectActions;
import org.jboss.system.ServiceMBeanSupport;
import org.jboss.util.CachePolicy;
import org.jboss.util.TimedCachePolicy;
import org.jboss.util.TimedCachePolicy.TimedEntry;

public class JaasSecurityManager extends ServiceMBeanSupport implements SubjectSecurityManager, RealmMapping {

   private String securityDomain;
   private CachePolicy domainCache;
   private CallbackHandler handler;
   private Method setSecurityInfo;
   protected Logger log;
   protected boolean trace;


   public JaasSecurityManager() {
      this("other", new SecurityAssociationHandler());
   }

   public JaasSecurityManager(String securityDomain, CallbackHandler handler) {
      this.securityDomain = securityDomain;
      this.handler = handler;
      String categoryName = this.getClass().getName() + '.' + securityDomain;
      this.log = Logger.getLogger(categoryName);
      this.trace = this.log.isTraceEnabled();
      Class[] sig = new Class[]{Principal.class, Object.class};

      try {
         this.setSecurityInfo = handler.getClass().getMethod("setSecurityInfo", sig);
      } catch (Exception var7) {
         String msg = "Failed to find setSecurityInfo(Princpal, Object) method in handler";
         throw new UndeclaredThrowableException(var7, msg);
      }

      this.log.debug("CallbackHandler: " + handler);
   }

   public void setCachePolicy(CachePolicy domainCache) {
      this.domainCache = domainCache;
      this.log.debug("CachePolicy set to: " + domainCache);
   }

   public void flushCache() {
      if(this.domainCache != null) {
         this.domainCache.flush();
      }

   }

   public String getSecurityDomain() {
      return this.securityDomain;
   }

   public Subject getActiveSubject() {
      return SecurityAssociation.getSubject();
   }

   public boolean isValid(Principal principal, Object credential) {
      return this.isValid(principal, credential, (Subject)null);
   }

   public boolean isValid(Principal principal, Object credential, Subject activeSubject) {
      JaasSecurityManager.DomainInfo cacheInfo = this.getCacheInfo(principal, true);
      if(this.trace) {
         this.log.trace("Begin isValid, principal:" + principal + ", cache info: " + cacheInfo);
      }

      boolean isValid = false;
      if(cacheInfo != null) {
         isValid = this.validateCache(cacheInfo, credential, activeSubject);
         if(cacheInfo != null) {
            cacheInfo.release();
         }
      }

      if(!isValid) {
         isValid = this.authenticate(principal, credential, activeSubject);
      }

      if(this.trace) {
         this.log.trace("End isValid, " + isValid);
      }

      return isValid;
   }

   public Principal getPrincipal(Principal principal) {
      Principal result = principal;
      CachePolicy var3 = this.domainCache;
      synchronized(this.domainCache) {
         JaasSecurityManager.DomainInfo info = this.getCacheInfo(principal, false);
         if(this.trace) {
            this.log.trace("getPrincipal, cache info: " + info);
         }

         if(info != null) {
            result = info.callerPrincipal;
            if(result == null) {
               result = principal;
            }

            info.release();
         }

         return result;
      }
   }

   public boolean doesUserHaveRole(Principal principal, Set rolePrincipals) {
      boolean hasRole = false;
      Subject subject = SubjectActions.getActiveSubject();
      if(subject != null) {
         if(this.trace) {
            this.log.trace("doesUserHaveRole(Set), subject: " + subject);
         }

         Group roles = this.getSubjectRoles(subject);
         if(this.trace) {
            this.log.trace("roles=" + roles);
         }

         if(roles != null) {
            Iterator iter = rolePrincipals.iterator();

            while(!hasRole && iter.hasNext()) {
               Principal role = (Principal)iter.next();
               hasRole = this.doesRoleGroupHaveRole(role, roles);
               if(this.trace) {
                  this.log.trace("hasRole(" + role + ")=" + hasRole);
               }
            }
         }

         if(this.trace) {
            this.log.trace("hasRole=" + hasRole);
         }
      }

      return hasRole;
   }

   public boolean doesUserHaveRole(Principal principal, Principal role) {
      boolean hasRole = false;
      Subject subject = SubjectActions.getActiveSubject();
      if(subject != null) {
         if(this.trace) {
            this.log.trace("doesUserHaveRole(Principal), subject: " + subject);
         }

         Group roles = this.getSubjectRoles(subject);
         if(roles != null) {
            hasRole = this.doesRoleGroupHaveRole(role, roles);
         }
      }

      return hasRole;
   }

   public Set getUserRoles(Principal principal) {
      HashSet userRoles = null;
      Subject subject = SubjectActions.getActiveSubject();
      if(subject != null) {
         if(this.trace) {
            this.log.trace("getUserRoles, subject: " + subject);
         }

         Group roles = this.getSubjectRoles(subject);
         if(roles != null) {
            userRoles = new HashSet();
            Enumeration members = roles.members();

            while(members.hasMoreElements()) {
               Principal role = (Principal)members.nextElement();
               userRoles.add(role);
            }
         }
      }

      return userRoles;
   }

   protected boolean doesRoleGroupHaveRole(Principal role, Group userRoles) {
      if(role instanceof NobodyPrincipal) {
         return false;
      } else {
         boolean isMember = userRoles.isMember(role);
         if(!isMember) {
            isMember = role instanceof AnybodyPrincipal;
         }

         return isMember;
      }
   }

   private boolean authenticate(Principal principal, Object credential, Subject theSubject) {
      Subject subject = null;
      boolean authenticated = false;
      LoginException authException = null;

      try {
         LoginContext e = this.defaultLogin(principal, credential);
         subject = e.getSubject();
         if(subject != null) {
            if(theSubject != null) {
               SubjectActions.copySubject(subject, theSubject);
            }

            authenticated = true;
            this.updateCache(e, subject, principal, credential);
         }
      } catch (LoginException var8) {
         if(principal != null && principal.getName() != null || this.trace) {
            this.log.trace("Login failure", var8);
         }

         authException = var8;
      }

      SubjectActions.setContextInfo("org.jboss.security.exception", authException);
      return authenticated;
   }

   private LoginContext defaultLogin(Principal principal, Object credential) throws LoginException {
      Object[] securityInfo = new Object[]{principal, credential};
      CallbackHandler theHandler = null;

      LoginException lc;
      try {
         theHandler = (CallbackHandler)this.handler.getClass().newInstance();
         this.setSecurityInfo.invoke(theHandler, securityInfo);
      } catch (Throwable var7) {
         if(this.trace) {
            this.log.trace("Failed to create/setSecurityInfo on handler", var7);
         }

         lc = new LoginException("Failed to setSecurityInfo on handler");
         lc.initCause(var7);
         throw lc;
      }

      Subject subject = new Subject();
      lc = null;
      if(this.trace) {
         this.log.trace("defaultLogin, principal=" + principal);
      }

      LoginContext lc1 = SubjectActions.createLoginContext(this.securityDomain, subject, theHandler);
      lc1.login();
      if(this.trace) {
         this.log.trace("defaultLogin, lc=" + lc1 + ", subject=" + SubjectActions.toString(subject));
      }

      return lc1;
   }

   private boolean validateCache(JaasSecurityManager.DomainInfo info, Object credential, Subject theSubject) {
      if(this.trace) {
         StringBuffer subjectCredential = new StringBuffer("Begin validateCache, info=");
         subjectCredential.append(info.toString());
         subjectCredential.append(";credential.class=");
         if(credential != null) {
            Class isValid = credential.getClass();
            subjectCredential.append(isValid.getName());
            subjectCredential.append('@');
            subjectCredential.append(System.identityHashCode(isValid));
         } else {
            subjectCredential.append("null");
         }

         this.log.trace(subjectCredential.toString());
      }

      Object subjectCredential1 = info.credential;
      boolean isValid1 = false;
      if(credential != null && subjectCredential1 != null) {
         if(subjectCredential1.getClass().isAssignableFrom(credential.getClass())) {
            if(subjectCredential1 instanceof Comparable) {
               Comparable a1 = (Comparable)subjectCredential1;
               isValid1 = a1.compareTo(credential) == 0;
            } else if(subjectCredential1 instanceof char[]) {
               char[] a11 = (char[])subjectCredential1;
               char[] a2 = (char[])credential;
               isValid1 = Arrays.equals(a11, a2);
            } else if(subjectCredential1 instanceof byte[]) {
               byte[] a12 = (byte[])subjectCredential1;
               byte[] a21 = (byte[])credential;
               isValid1 = Arrays.equals(a12, a21);
            } else if(subjectCredential1.getClass().isArray()) {
               Object[] a13 = (Object[])subjectCredential1;
               Object[] a22 = (Object[])credential;
               isValid1 = Arrays.equals(a13, a22);
            } else {
               isValid1 = subjectCredential1.equals(credential);
            }
         }
      } else {
         isValid1 = credential == null && subjectCredential1 == null;
      }

      if(isValid1 && theSubject != null) {
         SubjectActions.copySubject(info.subject, theSubject);
      }

      if(this.trace) {
         this.log.trace("End validateCache, isValid=" + isValid1);
      }

      return isValid1;
   }

   private JaasSecurityManager.DomainInfo getCacheInfo(Principal principal, boolean allowRefresh) {
      if(this.domainCache == null) {
         return null;
      } else {
         JaasSecurityManager.DomainInfo cacheInfo = null;
         CachePolicy var4 = this.domainCache;
         synchronized(this.domainCache) {
            if(allowRefresh) {
               cacheInfo = (JaasSecurityManager.DomainInfo)this.domainCache.get(principal);
            } else {
               cacheInfo = (JaasSecurityManager.DomainInfo)this.domainCache.peek(principal);
            }

            if(cacheInfo != null) {
               cacheInfo.acquire();
            }

            return cacheInfo;
         }
      }
   }

   private Subject updateCache(LoginContext lc, Subject subject, Principal principal, Object credential) {
      if(this.domainCache == null) {
         return subject;
      } else {
         long lifetime = 0L;
         if(this.domainCache instanceof TimedCachePolicy) {
            TimedCachePolicy info = (TimedCachePolicy)this.domainCache;
            lifetime = (long)info.getDefaultLifetime();
         }

         JaasSecurityManager.DomainInfo info1 = new JaasSecurityManager.DomainInfo(lifetime);
         info1.loginCtx = lc;
         info1.subject = new Subject();
         SubjectActions.copySubject(subject, info1.subject, true);
         info1.credential = credential;
         if(this.trace) {
            this.log.trace("updateCache, inputSubject=" + SubjectActions.toString(subject) + ", cacheSubject=" + SubjectActions.toString(info1.subject));
         }

         Set subjectGroups = subject.getPrincipals(Group.class);
         Iterator iter = subjectGroups.iterator();

         while(iter.hasNext()) {
            Group subjectPrincipals = (Group)iter.next();
            String p = subjectPrincipals.getName();
            if(p.equals("CallerPrincipal")) {
               Enumeration members = subjectPrincipals.members();
               if(members.hasMoreElements()) {
                  info1.callerPrincipal = (Principal)members.nextElement();
               }
            }
         }

         if(principal == null && info1.callerPrincipal == null) {
            Set subjectPrincipals1 = subject.getPrincipals(Principal.class);
            iter = subjectPrincipals1.iterator();

            while(iter.hasNext()) {
               Principal p1 = (Principal)iter.next();
               if(!(p1 instanceof Group)) {
                  info1.callerPrincipal = p1;
               }
            }
         }

         CachePolicy subjectPrincipals2 = this.domainCache;
         synchronized(this.domainCache) {
            if(this.domainCache.peek(principal) != null) {
               this.domainCache.remove(principal);
            }

            this.domainCache.insert(principal, info1);
            if(this.trace) {
               this.log.trace("Inserted cache info: " + info1);
            }
         }

         return info1.subject;
      }
   }

   private Group getSubjectRoles(Subject theSubject) {
      Set subjectGroups = theSubject.getPrincipals(Group.class);
      Iterator iter = subjectGroups.iterator();
      Group roles = null;

      while(iter.hasNext()) {
         Group grp = (Group)iter.next();
         String name = grp.getName();
         if(name.equals("Roles")) {
            roles = grp;
         }
      }

      return roles;
   }

   public static class DomainInfo implements TimedEntry {

      private static Logger log = Logger.getLogger(JaasSecurityManager.DomainInfo.class);
      private static boolean trace = log.isTraceEnabled();
      private LoginContext loginCtx;
      private Subject subject;
      private Object credential;
      private Principal callerPrincipal;
      private long expirationTime;
      private boolean needsDestroy;
      private int activeUsers;


      public DomainInfo(long lifetime) {
         this.expirationTime = lifetime;
         if(this.expirationTime != -1L) {
            this.expirationTime *= 1000L;
         }

      }

      synchronized int acquire() {
         ++this.activeUsers;
         return this.activeUsers;
      }

      synchronized int release() {
         --this.activeUsers;
         int users = this.activeUsers;
         if(this.needsDestroy && users == 0) {
            if(trace) {
               log.trace("needsDestroy is true, doing logout");
            }

            this.logout();
         }

         return users;
      }

      synchronized void logout() {
         if(trace) {
            log.trace("logout, subject=" + this.subject + ", this=" + this);
         }

         try {
            if(this.loginCtx != null) {
               this.loginCtx.logout();
            }
         } catch (Throwable var2) {
            if(trace) {
               log.trace("Cache entry logout failed", var2);
            }
         }

      }

      public void init(long now) {
         this.expirationTime += now;
      }

      public boolean isCurrent(long now) {
         boolean isCurrent = this.expirationTime == -1L;
         if(!isCurrent) {
            isCurrent = this.expirationTime > now;
         }

         return isCurrent;
      }

      public boolean refresh() {
         return false;
      }

      public void destroy() {
         if(trace) {
            log.trace("destroy, subject=" + this.subject + ", this=" + this + ", activeUsers=" + this.activeUsers);
         }

         synchronized(this) {
            if(this.activeUsers == 0) {
               this.logout();
            } else {
               if(trace) {
                  log.trace("destroy saw activeUsers=" + this.activeUsers);
               }

               this.needsDestroy = true;
            }

         }
      }

      public Object getValue() {
         return this;
      }

      public String toString() {
         StringBuffer tmp = new StringBuffer(super.toString());
         tmp.append('[');
         tmp.append(SubjectActions.toString(this.subject));
         tmp.append(",credential.class=");
         if(this.credential != null) {
            Class c = this.credential.getClass();
            tmp.append(c.getName());
            tmp.append('@');
            tmp.append(System.identityHashCode(c));
         } else {
            tmp.append("null");
         }

         tmp.append(",expirationTime=");
         tmp.append(this.expirationTime);
         tmp.append(']');
         return tmp.toString();
      }

   }
}
